//!Package InterSoft.found.types
// --------------------------------------------------------------------
// Copyright (c) 1995 InterSoft Co.  All Rights Reserved
// THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF InterSoft Co.
// The copyright notice above does not evidence any actual or intended 
// publication of such source code.
//
// $Id: parray.h,v 1.22 1996/05/31 15:26:15 valeria Exp $
//
// Description:
// Interface to generic Pointer Arrays
// --------------------------------------------------------------------

#ifndef	PARRAY_H
#define	PARRAY_H

#include <ifound.h>

#ifdef __GNUC__
#	pragma interface
#endif

#include <ifound/functype.h>
#include <ifound/template.h>

#define ARRAY_REALLOC	(8)

/**
 * Class BasePtrArray implements a generic array of pointers to void.
 * Its purpose is to allow working with arrays of pointers to any type 
 * of objects.
 * <em>No instances of BasePtrArray objects can be created</em>, it 
 * should be instantiated through derived classes, like <em>PtrArray
 * </em>.
 * @version $Revision: 1.22 $
 * @author $Author: valeria $
 * @see PtrArray
 */
class BasePtrArray {

	// to prevent its use
	void operator=(const BasePtrArray &);
	BasePtrArray(const BasePtrArray &);

protected:
	Int a_dim;
	Int chunkSize;
	void **c;

	BasePtrArray(Int n, Int nRealloc);
	~BasePtrArray();
	void * &elemTest(Int n);

public:

	Int dim() const;
	void reAlloc(Int n);
	Int add(const void *element);
	void del(Int n);
	Int del(const void *element);
	void ins(Int n, const void *element);
	bool isInRange(Int n) const;
	bool isValid(Int n);
	void clear();
	Int nextFree() const;
	void sort(IFPVCPVCP sortFunc);
	Int find(const void *element);
	int find(const void *el, IFPVCPVCP cmpFunc);

};

class BasePtrArrayScanner {
protected:

	void *const *base;
	void *const *curr;
	void *const *last;

	void scanUp_();
	void scanDown_();

public:
	BasePtrArrayScanner() {};

	void goToFirst()
		{ curr=base; scanUp_(); }
	void goToLast()
		{ curr=last-1; scanDown_(); }
	void goTo(Int offs)
		{ curr=base+offs; scanUp_(); }

	operator bool() const
		{ return (curr < last && curr >= base); }
	bool operator!() const
		{ return !bool(*this); }
	Int index()
		{ return curr-base; }

	void operator+=(Int offs)
		{ curr += offs; scanUp_(); }
	void operator-=(Int offs)
		{ curr -= offs; scanDown_(); }
	void operator++()
		{ curr++; scanUp_(); }
	void operator--()
		{ curr--; scanDown_(); }
/*	void *operator->()
		{ return *curr; }
	void *operator*()
		{ return *curr; }
*/
};

#define PtrArray(CONTENT)		name2(CONTENT,PtrArray)

/**
 * PtrArray is a class inherited from <em>BasePtrArray</em> 
 * which purpose is to behave like an array of any type of objects. 
 * Therefore it was implemented like a template class that must be 
 * <em>instantiated to a concrete object type before creating objects
 * </em>.<p> PtrArray was implemented like a macro definition, so the 
 * instantiation is slightly different from a normal class 
 * instantiation.
 * Before creating a PtrArray object the class must be declared in the
 * following way:
 * <pre>
 * declare(PtrArray, className)
 * </pre>
 * A new class PtrArray that holds elements of the class specified in
 * <em>className</em> is created. Then objects of that instance class 
 * can be instantiated.
 * This new class inherits from class BasePtrArray and its methods
 * call functions of this basE class, wich allows working with any kind
 * of pointers. For a detailed description of the member functions see 
 * class <em>BasePtrArray</em>. 
 * @version $Revision: 1.22 $
 * @author $Author: valeria $
 * @see BasePtrArray
 */
#define	PtrArraydeclare(CONTENT)									\
																	\
class PtrArray(CONTENT) : public BasePtrArray {						\
																	\
	typedef BasePtrArray inherited;									\
																	\
public :															\
	explicit PtrArray(CONTENT)(Int n = 0, Int chunk = ARRAY_REALLOC)\
			: BasePtrArray(n, chunk) {}								\
																	\
	CONTENT *&operator[](Int n)										\
		{ return (CONTENT * &) elemTest(n); }						\
																	\
	CONTENT *&elem(Int n) const										\
		{ assert(isInRange(n)); return (CONTENT * &) c[n]; }		\
																	\
	Int add(const CONTENT *element)									\
		{ return inherited::add(element); }							\
																	\
	Int del(const CONTENT *element)									\
		{ return inherited::del(element); }							\
																	\
	void del(Int n)													\
		{ inherited::del(n); }										\
																	\
	Int find(const CONTENT *element)								\
		{ return inherited::find(element); }						\
																	\
	void ins(Int n, const CONTENT *element)							\
		{ inherited::ins(n, element); }								\
																	\
	int find(const CONTENT *el, IFPVCPVCP cmpFunc)					\
		{ return inherited::find(el, cmpFunc); }					\
};

#define PtrArrayCursor(CONTENT)		name2(CONTENT,PtrArrayCursor)

struct GoToLast {};

#define PtrArrayCursordeclare(CONTENT)								\
																	\
class PtrArrayCursor(CONTENT) {										\
																	\
	CONTENT **base;													\
	CONTENT **curr;													\
	CONTENT **last;													\
																	\
public:																\
	explicit PtrArrayCursor(CONTENT)(PtrArray(CONTENT) &pa)			\
	{ assign(pa); }													\
	explicit PtrArrayCursor(CONTENT)(PtrArray(CONTENT) &pa, 		\
			const GoToLast&)										\
	{ assign(pa); goToLast(); }										\
	PtrArrayCursor(CONTENT)(PtrArray(CONTENT) &pa, Int from) 		\
	{ assign(pa,from); }											\
	PtrArrayCursor(CONTENT)(PtrArray(CONTENT) &pa, Int from,		\
			const GoToLast&)								 		\
	{ assign(pa,from); goToLast(); }								\
	PtrArrayCursor(CONTENT)(PtrArray(CONTENT) &pa, Int from, Int to)\
	{ assign(pa,from,to); }											\
	PtrArrayCursor(CONTENT)(PtrArray(CONTENT) &pa, Int from, Int to,\
			const GoToLast&)										\
	{ assign(pa,from,to); goToLast(); }								\
	PtrArrayCursor(CONTENT)()										\
	{ reset(); }													\
	PtrArrayCursor(CONTENT) &operator=(PtrArray(CONTENT) &pa)		\
	{ assign(pa); return *this; }									\
	void assign(PtrArray(CONTENT) &pa)								\
	{																\
		if (pa.dim() == 0)											\
			reset();												\
		else {														\
			base = curr = &pa.elem(0);								\
			last = base + pa.dim();									\
		}															\
	}																\
	void assign(PtrArray(CONTENT) &pa, Int from)					\
	{																\
		if (pa.dim() == 0)											\
			reset();												\
		else {														\
			base = &pa.elem(0);										\
			last = base + pa.dim();									\
			curr = base += from;									\
		}															\
	}																\
	void assign(PtrArray(CONTENT) &pa, Int from, Int to)			\
	{																\
		if (pa.dim() == 0)											\
			reset();												\
		else {														\
			base = &pa.elem(0);										\
			last = base + (to != INT_MAX ? to + 1 : pa.dim());		\
			curr = base += from;									\
		}															\
	}																\
	void reset()													\
	{ base = curr = last = NULL;}									\
	void goToFirst()												\
	{ curr = base; }												\
	void goToLast()													\
	{ if (base) curr = last - 1; }									\
	void goTo(Int offs)												\
	{ if (base) curr = base + offs; }								\
	Int index() const												\
	{ return curr - base; }											\
	operator bool()	const											\
	{ return (curr < last && curr >= base); }						\
	bool operator!() const											\
	{ return !bool(*this); }										\
	PtrArrayCursor(CONTENT) &operator+=(Int offs)					\
	{ if (base) curr += offs; return *this; }						\
	PtrArrayCursor(CONTENT) &operator-=(Int offs)					\
	{ if (base) curr -= offs; return *this; }						\
	PtrArrayCursor(CONTENT) &operator++()							\
	{ if (base) curr++; return *this; }								\
	PtrArrayCursor(CONTENT) &operator--()							\
	{ if (base) curr--; return *this; }								\
	CONTENT *operator->()											\
	{ return *curr; }												\
	CONTENT *&operator*()											\
	{ return *curr; }												\
};


#define PtrArrayScanner(CONTENT)     name2(CONTENT,PtrArrayScanner)

#define PtrArrayScannerdeclare(CONTENT)								\
																	\
class PtrArrayScanner(CONTENT) : public BasePtrArrayScanner {		\
public:																\
	PtrArrayScanner(CONTENT)()										\
		{ base = curr = last = NULL; }								\
	PtrArrayScanner(CONTENT) &operator=(PtrArray(CONTENT) &pa)		\
		{															\
			if ( pa.dim()==0 )										\
				base = curr = last = NULL;							\
			else													\
			{														\
				curr = base = (void*const*)&pa.elem(0);				\
				last = base + pa.dim();								\
				scanUp_();											\
			}														\
			return *this;											\
		}															\
	PtrArrayScanner(CONTENT)(PtrArray(CONTENT) &pa) 				\
		{															\
			if ( pa.dim()==0 )										\
				base = curr = last = NULL;							\
			else													\
			{														\
				curr = base = (void*const*)&pa.elem(0);				\
				last = base + pa.dim();								\
				scanUp_();											\
			}														\
		}															\
	PtrArrayScanner(CONTENT)(PtrArray(CONTENT) &pa, 				\
						Int from, Int to = INT_MAX)					\
		{															\
			if ( pa.dim()==0 )										\
				base = curr = last = NULL;							\
			else													\
			{														\
				base = (void*const*)&pa.elem(0);					\
				last = base + (to != INT_MAX ? to + 1 : pa.dim());	\
				curr = (base += from);								\
				scanUp_();											\
			}														\
		}															\
	CONTENT *operator->()											\
	{ return (CONTENT*)*curr; }										\
	CONTENT *operator*()											\
	{ return (CONTENT*)*curr; }										\
};

// Empty implementations
#define	PtrArrayimplement(CONTENT)
#define	PtrArrayCursorimplement(CONTENT)
#define	PtrArrayScannerimplement(CONTENT)

#include <ifound/parray.icc>

#endif // PARRAY_H

